@@ -18,8 +18,7 @@ from maggy.earlystop.abstractearlystop import AbstractEarlyStop
 
 
 class NoStoppingRule(AbstractEarlyStop):
-    """The no stopping rule never stops any trials early.
-    """
+    """The no stopping rule never stops any trials early."""
 
     @staticmethod
     def earlystop_check(to_check, finalized_trials, direction):
